# Program: libraryCopy.tcl
#
# This file performs the following functions while building the freeWrap application:
#     1) Copy the TCL/TK standard scripts to the specified destination
#     2) Add the standard TCL library directories to the tcl_pkgPath variable
#     3) Generate pkgIndex.tcl files so that TCL modules can be loaded with [package require]
#
# Revison  Date           Author             Description
# -------  -------------  -----------------  ----------------------------------------------
#   6.4    Jan. 20, 2008  Dennis R. LaBelle  1) This file replaced the shrink.tcl script.
#                                               Extra white space is no longer removed from
#                                               the copied TCL scripts.
#
#   6.41   Feb. 17, 2008  Dennis R. LaBelle  1) Modified code so that TCL packge paths are
#                                               added to tcl_pkgPath variable instead of
#                                               auto_path.
#                                            2) Added code to copy the TCL module directories.
#                                            3) Added code to generate pkgIndex.tcl files
#                                               so that TCL modules can be loaded with a
#                                               [package require] command.

proc main {} {
global argv

set mountpt {}
set pkg [lindex $argv 0]
set scriptdir [file dirname [lindex $argv 1]]
set zipdir [file dirname [lindex $argv 2]]
set basedir [file dirname $zipdir]
set baselen [string length $basedir]
set scriptTMdir [file rootname $scriptdir]
set zipTMdir [file join [file dirname $zipdir] [file tail $scriptTMdir]]
set autolist {}

switch $pkg {
	tcl	{
		  file delete -force $zipdir
		  file copy -force $scriptdir $zipdir
		  file delete -force $zipTMdir
		  file copy -force $scriptTMdir $zipTMdir

		  # Get list of subdirectories copied
		  set autolist {}
		  foreach branch "$zipdir $zipTMdir" {
			  set dirlist [glob -nocomplain -directory $branch -types d *]
			  foreach newdir $dirlist {
				  lappend autolist [string range $newdir $baselen end]
				  if {[string length [glob -nocomplain $newdir/*.tm]]} {
				      pkg_mkIndex $newdir *.tm
				     }
				}
			  while {[llength $dirlist]} {
				 set newdir [lindex $dirlist 0]
				 set newlist [glob -nocomplain -types d [file join $newdir *]]
				 foreach newdir $newlist {
					 lappend autolist [string range $newdir $baselen end]
					 if {[string length [glob -nocomplain $newdir/*.tm]]} {
					     pkg_mkIndex $newdir *.tm
					    }
				       }
				 set autolist [concat $autolist $newlist]
				 set dirlist [concat [lrange $dirlist 1 end] $newlist]
				}
			}

		  # Add list of subdirectories to tcl_pkgPath in init.tcl file.
		  set initfile [file join $zipdir init.tcl]
		  file rename -force $initfile ${initfile}.org
		  set fout2 [open $initfile w]
		  puts $fout2 {global tcl_pkgPath} 
		  puts $fout2 {set tcl_pkgPath {}}
		  foreach dir $autolist {
			  set newpath [file join $mountpt/tcl $dir]
			  puts $fout2 "lappend tcl_pkgPath $newpath"
			}
		  set fin2 [open ${initfile}.org r]
		  fcopy $fin2 $fout2
		  close $fin2
		  close $fout2
		  file delete -force ${initfile}.org
		}
	tk	{
		  file delete -force $zipdir
		  file copy -force $scriptdir $zipdir

		  # Remove unnecessary subdirectories.
		  file delete -force [file join $zipdir demos]
		}
	blt	{
		  file delete -force $zipdir
		  file copy $scriptdir $zipdir

		  # Remove unnecessary subdirectories.
		  file delete -force [file join $zipdir demos]
		  file delete -force [file join $zipdir html]
		}
	default	{
		  exit 1
		}
     }
}

main
exit 0